{% extends "basic/layout.html" %}

{% set theme_css_files = []%}
{% if theme_bootstrap_version == "3" %}
  {% set bootstrap_version, bootstrap_additional_css, navbar_version = "3.1.0", "theme", "" %}
  {% set bs_span_prefix = "col-md-" %}
{% else %}
  {% set bootstrap_version, bootstrap_additional_css, navbar_version = "2.3.2", "responsive", "-2" %}
  {% set bs_span_prefix = "span" %}
{% endif %}

{% if theme_bootswatch_theme %}
  {# BS2 needs "bootstrap-responsive.css". BS3 doesn't. #}
  {% if theme_bootstrap_version == "3" %}
    {% set theme_css_files = theme_css_files + [
        '_static/bootswatch-' + bootstrap_version + '/' + theme_bootswatch_theme + '/bootstrap.min.css',
        '_static/bootstrap-sphinx.css'
      ]
    %}
  {% else %}
    {% set theme_css_files = theme_css_files + [
        '_static/bootswatch-' + bootstrap_version + '/' + theme_bootswatch_theme + '/bootstrap.min.css',
        '_static/bootstrap-' + bootstrap_version + '/css/bootstrap-' + bootstrap_additional_css + '.min.css',
        '_static/bootstrap-sphinx.css'
      ]
    %}
  {% endif %}
{% else %}
  {% set theme_css_files = theme_css_files + ['_static/main.min.css', '_static/styles.css']
  %}
{% endif %}

{% if not bootswatch_css_custom %}
  {% set bootswatch_css_custom = [] %}
{% endif %}
{% set css_files = css_files + theme_css_files + bootswatch_css_custom %}

{% set script_files = script_files + [
    '_static/js/jquery-1.11.0.min.js',
    '_static/js/jquery-fix.js',
    '_static/bootstrap-' + bootstrap_version + '/js/bootstrap.min.js',
    '_static/bootstrap-sphinx.js'
  ]
%}

{%- set render_sidebar = (not embedded) and (not theme_nosidebar|tobool) and sidebars %}

{%- set bs_content_width = render_sidebar and "9" or "12"%}

{%- block doctype -%}
<!DOCTYPE html>
{%- endblock %}

{# Sidebar: Rework into our Bootstrap nav section. #}
{% macro navBar() %}
<header class="banner navbar navbar-default navbar-static-top" role="banner">
  <div class="container">
    <div class="navbar-header">
      <button type="button" class="navbar-toggle" data-toggle="collapse" data-target=".navbar-collapse">
        <span class="sr-only">Toggle navigation</span>
        <span class="icon-bar"></span>
        <span class="icon-bar"></span>
        <span class="icon-bar"></span>
      </button>
       <a class="brand" href="http://nextcloud.com/" rel="home" title="Nextcloud.com"><img class="logo" src="{{ pathto('_static/img/logo.svg', 1) }}" itemprop="logo"></a>
    </div>

    <nav class="collapse navbar-collapse" role="navigation">
      <ul id="menu-header" class="nav navbar-nav">
          <li><a target="_blank" href="https://demo.nextcloud.com">Demo</a></li>
          <li><a href="https://nextcloud.com/news/">News</a></li>
          <li><a href="https://nextcloud.com/features/">Features</a></li>
          <li><a href="https://nextcloud.com/about/">About us</a></li>
          <li><a href="https://nextcloud.com/contribute/">Get involved</a></li>
          <li><a href="https://nextcloud.com/support/">Support</a></li>
          <li><a href="https://nextcloud.com/enterprise/">Enterprise</a></li>
          <li class="menu-install"><a href="https://nextcloud.com/install/">Download</a></li>
      </ul>
    </nav>
  </div>
</header>
{% endmacro %}

{% if theme_bootstrap_version == "3" %}
  {%- macro bsidebar() %}
      {%- if render_sidebar %}
      <div class="{{ bs_span_prefix }}3">
        <div id="sidebar" class="bs-sidenav" role="complementary">
          {%- for sidebartemplate in sidebars %}
            {%- include sidebartemplate %}
          {%- endfor %}
        </div>
      </div>
      {%- endif %}
  {%- endmacro %}
{% else %}
  {%- macro bsidebar() %}
      {%- if render_sidebar %}
      <div class="{{ bs_span_prefix }}3">
        <div id="sidebar" class="bs-sidenav well" data-spy="affix">
          {%- for sidebartemplate in sidebars %}
            {%- include sidebartemplate %}
          {%- endfor %}
        </div>
      </div>
      {%- endif %}
  {%- endmacro %}
{% endif %}

{%- block extrahead %}
<meta charset='utf-8'>
<meta http-equiv='X-UA-Compatible' content='IE=edge,chrome=1'>
<meta name='viewport' content='width=device-width, initial-scale=1.0, maximum-scale=1'>
<meta name="apple-mobile-web-app-capable" content="yes">
<meta name="theme-color" content="#1d2d44">
<link rel="shortcut icon" href="https://nextcloud.com/wp-content/themes/next/assets/img/common/favicon.png" />
<link rel="apple-touch-icon-precomposed" href="https://nextcloud.com/wp-content/themes/next/assets/img/common/favicon.png" />
<!-- Piwik -->
<script type="text/javascript">
    var _paq = _paq || [];
    _paq.push(['trackPageView']);
    _paq.push(['enableLinkTracking']);
    (function() {
        var u="//stats.nextcloud.com/";
        _paq.push(['setTrackerUrl', u+'piwik.php']);
        _paq.push(['setSiteId', 3]);
        var d=document, g=d.createElement('script'), s=d.getElementsByTagName('script')[0];
        g.type='text/javascript'; g.async=true; g.defer=true; g.src=u+'piwik.js'; s.parentNode.insertBefore(g,s);
    })();
</script>
<noscript><p><img src="//stats.nextcloud.com/piwik.php?idsite=3" style="border:0;" alt="" /></p></noscript>
<!-- End Piwik Code -->
{% endblock %}

{# Silence the sidebar's, relbar's #}
{% block header %}{% endblock %}
{% block relbar1 %}{% endblock %}
{% block relbar2 %}{% endblock %}
{% block sidebarsourcelink %}{% endblock %}

{%- block content %}
{{ navBar() }}
<div class="wrap container not-front">
  <div class="content row">
  <main class="main">
    <div class="row page-content-header">
      <div class="col-md-8">
        <h1><a href="{{ pathto(master_doc) }}">{{ project|e }}</a></h1>
        <div class="doc-version-switch">
          Versions: 
        {%- for v in doc_versions %}
          <a href="https://docs.nextcloud.com/server/{{ v }}/{{current_doc}}/{{pagename}}.html" {% if version == v %} class="current-doc"{% endif %}>{{v}}</a>
        {%- endfor %}
        </div>
      </div>
      <div class="col-md-4">
      {% if theme_show_search %}
        <form class="headersearch" style="margin: 16px 0;" action="{{ pathto('search') }}" method="get">
        <input type="text" value="" name="q" id="q" class="form-control" /> 
        <button  class="btn btn-default" type="submit" id="searchsubmit">Search</button>
        </form>
      {% endif %}
      </div>
    </div>
    {% block sidebartoc %}
			<div class="row">
				<div class="col-md-3">
					<div class="sidebar">
							<div class="menu-support-container">
								<ul id="menu-support" class="menu">
									<ul>
										
<li><a href="{{ pathto(master_doc) }}">Table of Contents</a></li>
									</ul>
                  {{ toctree(maxdepth=3) }}
								</ul>
							</div>
					</div>
				</div>
        {% endblock %}

				<div class="col-md-9">
					<div class="page-content">
            {% include "relations.html" %}
						{% block body %}{% endblock %}
            {% include "relations.html" %}
					</div>
				</div>
			</div>
  </main>  
  </div>
</div>
{%- endblock %}

{%- block footer %}
<div class="container">
<div class="row">
    <div class="col-lg-12 footer-social-icons">
      <p class="text-center"><a target="_blank" href="https://plus.google.com/104036748063781940910/about"><img width=50 src="{{ pathto('_static/img/social/googleplus.svg', 1) }}" title="Follow us on Google Plus!" alt="Follow us on Google Plus!"></img></a>
      <a target="_blank" href="https://www.facebook.com/Nextclouders"><img width=50 src="{{ pathto('_static/img/social/facebook.svg', 1) }}" title="Like our Facebook page!" alt="Like our Facebook page!"></img></a>
      <a target="_blank" href="https://twitter.com/Nextclouders"><img width=50 src="{{ pathto('_static/img/social/twitter.svg', 1) }}" title="Subscribe to our Twitter channel!" alt="Subscribe to our Twitter channel!"></img></a>
      <a target="_blank" href="https://nextcloud.com/blogfeed"><img class="img-circle" width=50 src="{{ pathto('_static/img/social/rss.svg', 1) }}" title="Subscribe to our news feed!" alt="Subscribe to our news feed!"></img></a>
      <a target="_blank" href="https://newsletter.nextcloud.com/"><img class="img-circle" width=50 src="{{ pathto('_static/img/social/mail.svg', 1) }}"  title="Subscribe to our newsletter!" alt="Subscribe to our newsletter!"></img></a></p>
    </div>
    <div class="text-center">
      <p>All documentation licensed under the <a href="https://creativecommons.org/licenses/by/3.0/us/">Creative Commons Attribution 3.0 Unported license</a>.</p>
      <p><a href="https://github.com/nextcloud/documentation/graphs/contributors">See who contributed to our documentation/credits</a>.</p>
    </div>
</div>
</div>
<footer class="content-info" role="contentinfo">
<div class="container">
   <div class="row">
     <div class="col-sm-2 col-sm-offset-1">
        <div class="footer-nav">
          <h4>About Nextcloud</h4>
          <ul id="menu-about" class="menu">
            <li><a href="https://nextcloud.com/about/">About us</a></li>
            <li><a href="https://nextcloud.com/contact/">Contact us</a></li>
            <li><a href="https://nextcloud.com/press/">Press</a></li>
            <li><a href="https://nextcloud.com/privacy/">Privacy policy</a></li>
          </ul>
        </div>
     </div>
     <div class="col-sm-2 col-sm-offset-2">
     </div>
     <div class="col-sm-2 col-sm-offset-2">
        <div class="footer-nav">
          <h4>Interact</h4>
          <ul id="menu-interact" class="menu">
            <li class="menu-irc-channel"><a href="https://webchat.freenode.net/?channels=nextcloud">IRC Channel</a></li>
            <li class="menu-forums"><a href="https://help.nextcloud.com/">Forums</a></li>
          </ul>
        </div>
     </div>
   </div>
   <div class="row">
      <div class="col-lg-12 footer-text">
        <p>
          {%- if show_copyright %}
            {%- if hasdoc('copyright') %}
              {% trans path=pathto('copyright'), copyright=copyright|e %}&copy; <a href="{{ path }}">Copyright</a> {{ copyright }}.{% endtrans %}<br/>
            {%- else %}
              {% trans copyright=copyright|e %}&copy; Copyright {{ copyright }}.{% endtrans %}  <small><a href="https://nextcloud.com/impressum">Legal Notice</a></small><br/>
            {%- endif %}
          {%- endif %}
          {%- if last_updated %}
            {% trans last_updated=last_updated|e %}Last updated on {{ last_updated }}.{% endtrans %}<br/>
          {%- endif %}
          {%- if show_sphinx %}
            {% trans sphinx_version=sphinx_version|e %}Created using <a href="http://sphinx.pocoo.org/">Sphinx</a> {{ sphinx_version }}.{% endtrans %}<br/>
          {%- endif %}
        </p>
      </div>
    </div>
  </div>
</footer>
{%- endblock %}
