//{{NO_DEPENDENCIES}}
// Microsoft Visual C++ generated include file.
// Used by NCOverlay.rc
//
#define IDI_STATE_ERROR                 110
#define IDI_STATE_ERROR_SWM             111
#define IDI_STATE_OK                    112
#define IDI_STATE_OK_SWM                113
#define IDI_STATE_SYNC                  114
#define IDI_STATE_SYNC_SWM              115
#define IDI_STATE_WARNING               116
#define IDI_STATE_WARNING_SWM           117

// Next default values for new objects
// 
#ifdef APSTUDIO_INVOKED
#ifndef APSTUDIO_READONLY_SYMBOLS
#define _APS_NEXT_RESOURCE_VALUE        118
#define _APS_NEXT_COMMAND_VALUE         40001
#define _APS_NEXT_CONTROL_VALUE         1001
#define _APS_NEXT_SYMED_VALUE           101
#endif
#endif
