//
//  NCDesktopClientSocketKit.h
//  NCDesktopClientSocketKit
//
//  Created by Claudio Cambra on 23/12/22.
//

#import <Foundation/Foundation.h>

//! Project version number for NCDesktopClientSocketKit.
FOUNDATION_EXPORT double NCDesktopClientSocketKitVersionNumber;

//! Project version string for NCDesktopClientSocketKit.
FOUNDATION_EXPORT const unsigned char NCDesktopClientSocketKitVersionString[];

// In this header, you should import all the public headers of your framework using statements like #import <NCDesktopClientSocketKit/PublicHeader.h>

#import <NCDesktopClientSocketKit/LocalSocketClient.h>
#import <NCDesktopClientSocketKit/LineProcessor.h>
